/*!
 ************************************************************************
 * \file adaptive_quantization.c
 *
 * \brief
 *    adaptive quantization matrix selection routines
 *
 * \author
 *  Akiyuki Tanizawa       <akiyuki.tanizawa@toshiba.co.jp>           \n
 *  Takeshi Chujoh         <takeshi.chujoh@toshiba.co.jp>             \n
 *
 * This software may be used for research purposes only.
 * TOSHIBA would also appreciate that any technical report, conference or
 * journal paper which uses the software includes one of the following references:
 *    - A. Tanizawa, T. Chujoh, "Adaptive Quantization Matrix Selection",
 *                              VCEG contribution D.266, Geneva, November 2006.
 *    - A. Tanizawa, T. Chujoh, "Simulation results of AQMS on KTA software",
 *                              VCEG contribution VCEG-AC07, Klagenfurt, July 2006.
 *    - A. Tanizawa, T. Chujoh, "Adaptive Quantization Matrix Selection on KTA software",
 *                              VCEG contribution VCEG-AD06, Hangzhou, Oct. 2006.
 *    - A. Tanizawa, T. Chujoh, "Simulation results of AQMS on KTA software version 1.3",
 *                              VCEG contribution VCEG-AF08, San Jose, April 2007.
 ************************************************************************
 */

/*
 * CONFIGURATION:
 *
 * Configuration of the .Cfg file:
 * In section KTA STUFF:
 *
 * // AQMS command:
 * UseAdaptiveQuantMatrix		= 0 (0:disable, 1:enable)
 * // Adaptive QP command:
 * UseAdaptiveQuantParam		= 0 (0:disable, 1:enable)
 *
 * NOTE:
 * Both AQMS and Adaptive QP can perform on VCEG-AA10 conditions.
 *    - TK. Tan, G. Sullivan, T. Wedi, "Recommended Simulation Common Conditions for Coding Efficiency Experiments",
 *                              VCEG contribution VCEG-AA10, Nice, Oct. 2005.
 *    - TK. Tan, G. Sullivan, T. Wedi, "Recommended Simulation Common Conditions for Coding Efficiency Experiments",
 *                              VCEG contribution VCEG-AE10r1, Marrakech, Jan. 2007.
 ********************
 * KNOWN LIMITATIONS
 ********************
 * Not compatible with:                
 *	- interlace coding
 *	- RdOpt != 1 	  
 *  - RateControl
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <string.h>
#include "global.h"
#include "parset.h"
#include "adaptive_quantization.h"
#include "memalloc.h"
#include "malloc.h"
#include "defines.h"
#include "mbuffer.h"
#include "image.h"
#include "rdopt_coding_state.h"

#ifdef ADAPTIVE_FD_SD_CODING
#include "cabac.h"
#endif

#if (defined(ADAPTIVE_QUANTIZATION)||defined(ADAPTIVE_QP))

//------- for image.c
extern pic_parameter_set_rbsp_t *PicParSet[MAXPPS];
extern void copy_params();
extern void code_a_picture(Picture *pic);
extern void find_distortion();
extern void frame_picture (Picture *frame, int method);
//------- end

//------- for mode_decision.c
extern CSptr cs_mb, cs_b8, cs_cm, cs_imb, cs_ib8, cs_ib4, cs_pc;
extern void  store_coding_state   (CSptr);
extern const int  b8_mode_table[6];
extern void list_prediction_cost(int list, int block, int mode, RD_PARAMS enc_mb, int bmcost[5], char best_ref[2]);
extern void determine_prediction_list(int, int [5], char [2], short *, int *, short *);
extern int    RDCost_for_macroblocks (double, int, double*, double*, int);
extern double RDCost_for_8x8blocks (int*, int64*, double, int, int, short, short, short);
extern void compute_mode_RD_cost(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcostD, double *min_rateD, int i16mode, short bslice, short *inter_skip);
extern void compute_mode_RD_cost_Final(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcostD, double *min_rateD, int i16mode, short bslice, short *inter_skip);
extern RD_8x8DATA tr4x4, tr8x8;
extern void   StoreNewMotionVectorsBlock8x8(int, int, int, int, int, int, int);
extern void   SetRefAndMotionVectors (int, int, int, int, int);
extern void   SetModesAndRefframeForBlocks (int);
extern void   store_macroblock_parameters (int);
extern int    ****cofAC_8x8ts;        // [8x8block][4x4block][level/run][scan_pos]
extern int    ****cofAC, ****cofAC8x8;        // [8x8block][4x4block][level/run][scan_pos]
extern int    cbp_blk8x8;
extern int    cbp, cbp8x8, cnt_nonz_8x8;
extern int64  cbp_blk;
extern int64  cbp_blk8_8x8ts;
extern int    cbp8_8x8ts;
extern int    cnt_nonz8_8x8ts;
extern int    best_c_imode;
extern short  best_mode;
extern short  best8x8mode          [4]; // [block]
extern short  best8x8pdir [MAXMODE][4]; // [mode][block]
extern short  best8x8fwref[MAXMODE][4]; // [mode][block]
extern short  best8x8bwref[MAXMODE][4]; // [mode][block]
//------- end

/*!
*************************************************************************************
* \brief
*    Reset best coding parameter
*************************************************************************************
*/
void PreSetBestCodingParameters(int best_mode, Macroblock* currMB, RD_PARAMS enc_mb, short bslice, short *inter_skip, int best_ctr16x16)
{
  //  int   bmcost[5] = {INT_MAX}, 
  int   block;
  //  char  best_ref[2] = {0, -1};
  int   cost_direct=0, have_direct=0, cost=0, cost8x8_direct=0;
  int   pslice = (img->type==P_SLICE);
   
  if( best_mode <= P8x8)
  {
    //===== set direct motion vectors =====
    if (bslice)
    {
      Get_Direct_Motion_Vectors ();
    }
    
    if (best_mode==1 && bslice)
    {
      best8x8pdir[1][0] = best8x8pdir[1][1] = best8x8pdir[1][2] = best8x8pdir[1][3] = best_ctr16x16;
    }
    
    if ((!(*inter_skip)) && enc_mb.valid[P8x8])
    {
      giRDOpt_B8OnlyFlag = 1;
	
      tr8x8.cost8x8 = INT_MAX;
      tr4x4.cost8x8 = INT_MAX;
      //===== store coding state of macroblock =====
      store_coding_state (cs_mb);
      
      currMB->all_blk_8x8 = -1;
      
      if (input->Transform8x8Mode)
      {  
        tr8x8.cost8x8 = 0;
        //===========================================================
        // Check 8x8 partition with transform size 8x8 
        //===========================================================
        //=====  LOOP OVER 8x8 SUB-PARTITIONS  (Motion Estimation & Mode Decision) =====
        for (cost_direct=cbp8x8=cbp_blk8x8=cnt_nonz_8x8=0, block=0; block<4; block++)
        {
          SubmacroblockCalcCost(enc_mb, &tr8x8, currMB, cofAC_8x8ts[block],
          &have_direct, bslice, block, &cost_direct, &cost, &cost8x8_direct, 1);
          best8x8mode       [block] = tr8x8.part8x8mode [block];
          best8x8pdir [P8x8][block] = tr8x8.part8x8pdir [block];
          best8x8fwref[P8x8][block] = tr8x8.part8x8fwref[block];
          best8x8bwref[P8x8][block] = tr8x8.part8x8bwref[block];
	}
        
        // following params could be added in RD_8x8DATA structure
        cbp8_8x8ts      = cbp8x8;
        cbp_blk8_8x8ts  = cbp_blk8x8;
        cnt_nonz8_8x8ts = cnt_nonz_8x8;
        currMB->luma_transform_size_8x8_flag = 0; //switch to 4x4 transform size

        //--- re-set coding state (as it was before 8x8 block coding) ---
        //reset_coding_state (cs_mb);        
      }// if (input->Transform8x8Mode)
      

      if (input->Transform8x8Mode != 2)  
      {
        tr4x4.cost8x8 = 0;
        //=================================================================
        // Check 8x8, 8x4, 4x8 and 4x4 partitions with transform size 4x4
        //=================================================================
        //=====  LOOP OVER 8x8 SUB-PARTITIONS  (Motion Estimation & Mode Decision) =====
        for (cost_direct=cbp8x8=cbp_blk8x8=cnt_nonz_8x8=0, block=0; block<4; block++)
        {
          SubmacroblockCalcCost(enc_mb, &tr4x4, currMB, cofAC8x8[block],
            &have_direct, bslice, block, &cost_direct, &cost, &cost8x8_direct, 0);
          
          best8x8mode       [block] = tr4x4.part8x8mode [block];
          best8x8pdir [P8x8][block] = tr4x4.part8x8pdir [block];
          best8x8fwref[P8x8][block] = tr4x4.part8x8fwref[block];
          best8x8bwref[P8x8][block] = tr4x4.part8x8bwref[block];
        }          
        //--- re-set coding state (as it was before 8x8 block coding) ---
        // reset_coding_state (cs_mb);  
      }// if (input->Transform8x8Mode != 2)
      
      //--- re-set coding state (as it was before 8x8 block coding) ---
      reset_coding_state (cs_mb);
      
      giRDOpt_B8OnlyFlag = 0;
    }
    else // if (enc_mb.valid[P8x8])
    {
      tr4x4.cost8x8 = INT_MAX;
    }
    
    // Find a motion vector for the Skip mode
    if(input->rdopt != 2 && pslice)
      FindSkipModeMotionVector ();

  }
}


/*!
*************************************************************************************
* \brief
*    Mode Decision for an 8x8 sub-macroblock
*************************************************************************************
*/
void SubmacroblockCalcCost(RD_PARAMS enc_mb, 
                                 RD_8x8DATA *dataTr, 
                                 Macroblock *currMB,
                                 int ***cofACtr, 
                                 int *have_direct, 
                                 short bslice, 
                                 int block, 
                                 int *cost_direct,                                  
                                 int *cost,
                                 int *cost8x8_direct,
                                 int transform8x8)  
{
  int j0, i0, j1, i1;
  int i,j, k;
  int min_cost8x8, index;
  double min_rdcost, rdcost = 0.0;
  short best_pdir = 0;
  char best_ref[2] = {0, -1};
  int mode;
  int64 curr_cbp_blk;
  int bmcost[5] = {INT_MAX};
  int cnt_nonz = 0;
  short bi_pred_me;
  int best_cnt_nonz = 0;
  int maxindex =  (transform8x8) ? 2 : 5;
  int pix_x, pix_y;
  int block_x, block_y;
  
  int fadjust[16][16], fadjustCr[2][16][16];    
#ifdef ADAPTIVE_FD_SD_CODING
  float adjust_adaptive_f_spatial_domain_4x4=0.0;
  float adjust_adaptive_f_spatial_domain_8x8=0.0;
#endif
   
  int (*fadjustTransform)[16][16] = transform8x8? img->fadjust8x8 : img->fadjust4x4;
  int (*fadjustTransformCr)[2][16][16] = transform8x8? img->fadjust8x8Cr : img->fadjust4x4Cr;
  int lumaAdjustIndex = transform8x8? 2 : 3;
  int chromaAdjustIndex = transform8x8? 0 : 2;
  int fadjust_aqms[2][16][16], fadjustCr_aqms[2][2][16][16];    
#ifdef BEST_NZ_COEFF
  int best_nz_coeff[2][2];
#endif

  //--- set coordinates ---
  j0 = ((block/2)<<3);    j1 = (j0>>2);
  i0 = ((block%2)<<3);    i1 = (i0>>2);

#ifdef BEST_NZ_COEFF
  for(j = 0; j <= 1; j++)
  {
    for(i = 0; i <= 1; i++)
      best_nz_coeff[i][j] = img->nz_coeff[img->current_mb_nr][i1 + i][j1 + j] = 0;
  }
#endif

  if(gdwFinalEncodeFlag) pAQMSParam = gpBestAdaptiveQuantParam;
  else                   pAQMSParam = gpAdaptiveQuantParam;

  if (transform8x8)
    currMB->luma_transform_size_8x8_flag = 1; //switch to transform size 8x8
  
  //--- store coding state before coding ---
  store_coding_state (cs_cm);

  //=====  LOOP OVER POSSIBLE CODING MODES FOR 8x8 SUB-PARTITION  =====
  for (min_cost8x8=INT_MAX, min_rdcost=1e30, index=(bslice?0:1); index<maxindex; index++)
  {
    mode = b8_mode_table[index]; 
    if(mode != dataTr->part8x8mode [block]) continue;

    *cost = 0;
    if (enc_mb.valid[mode] && (transform8x8 == 0 || mode != 0 || (mode == 0 && active_sps->direct_8x8_inference_flag)))
    {
      curr_cbp_blk = 0;

      if(gdwFinalEncodeFlag && ( img->type == B_SLICE || (img->type == P_SLICE) ) )
      {
        if (mode==0)
        {
          //--- Direct Mode ---                                      
          block_x = img->block_x+(block&1)*2;
          block_y = img->block_y+(block&2);
          best_ref[LIST_0] = direct_ref_idx[LIST_0][block_y][block_x];
          best_ref[LIST_1] = direct_ref_idx[LIST_1][block_y][block_x];
          best_pdir   = direct_pdir[block_y][block_x];
        } // if (mode==0)
        else
        {        
          //======= motion estimation for all reference frames ========
          //-----------------------------------------------------------
          PartitionMotionSearch (mode, block, enc_mb.lambda_mf);
        
          //--- get cost and reference frame for LIST 0 prediction ---
          bmcost[LIST_0] = INT_MAX;
          list_prediction_cost(LIST_0, block, mode, enc_mb, bmcost, best_ref);
        
          //store LIST 0 reference index for every block
          block_x = img->block_x+(block&1)*2;
          block_y = img->block_y+(block&2);
          for (j = block_y; j< block_y + 2; j++)
          {
            for (i = block_x; i < block_x + 2; i++)
            {
              enc_picture->ref_idx   [LIST_0][j][i] = best_ref[LIST_0];
              enc_picture->ref_pic_id[LIST_0][j][i] = 
                enc_picture->ref_pic_num[enc_mb.list_offset[LIST_0]][(short)best_ref[LIST_0]];
            }
          } 
        
          if (bslice)
          {
            //--- get cost and reference frame for LIST 1 prediction ---
            bmcost[LIST_1] = INT_MAX;
            bmcost[BI_PRED] = INT_MAX;
            list_prediction_cost(LIST_1, block, mode, enc_mb, bmcost, best_ref);
          
            // Compute bipredictive cost between best list 0 and best list 1 references
            list_prediction_cost(BI_PRED, block, mode, enc_mb, bmcost, best_ref);
          
            //--- get prediction direction ----          
            determine_prediction_list(mode, bmcost, best_ref, &best_pdir, cost, &bi_pred_me);
          
            //store backward reference index for every block
            for (j = block_y; j< block_y + 2; j++)
            {
              for (i = block_x; i < block_x + 2; i++)
              {
                enc_picture->ref_idx[LIST_0][j][i] = best_ref[LIST_0];
                enc_picture->ref_idx[LIST_1][j][i] = best_ref[LIST_1];
              }
            }
          } // if (bslice)
          else
          {
            best_pdir = 0;
            *cost     = bmcost[LIST_0];
          }
        } // if (mode!=0)
      }

       best_pdir        = dataTr->part8x8pdir [block];
       best_ref[LIST_0] = dataTr->part8x8fwref[block];
       best_ref[LIST_1] = dataTr->part8x8bwref[block];

      if (input->rdopt)
      {
        //--- get and check rate-distortion cost ---
        rdcost = RDCost_for_8x8blocks (&cnt_nonz, &curr_cbp_blk, enc_mb.lambda_md,
          block, mode, best_pdir, best_ref[LIST_0], best_ref[LIST_1]);
      }
      else
      {
        if (*cost!=INT_MAX)
          *cost += (REF_COST (enc_mb.lambda_mf, B8Mode2Value (mode, best_pdir),
          enc_mb.list_offset[(best_pdir<1?LIST_0:LIST_1)]) - 1);
      }
      //--- set variables if best mode has changed ---
      if ( ( input->rdopt && rdcost < min_rdcost) 
        || (!input->rdopt && *cost < min_cost8x8))
      {
        min_cost8x8                 = *cost;
        min_rdcost                  = rdcost;
        dataTr->part8x8mode [block] = mode;
        dataTr->part8x8pdir [block] = best_pdir;
        dataTr->part8x8fwref[block] = best_ref[LIST_0];
        dataTr->part8x8bwref[block] = best_ref[LIST_1];
      
        img->mb_data[img->current_mb_nr].b8mode[block] = mode;
        
#ifdef BEST_NZ_COEFF
        for(j = 0; j <= 1; j++)
        {
          for(i = 0; i <= 1; i++)
            best_nz_coeff[i][j]= cnt_nonz ? img->nz_coeff[img->current_mb_nr][i1 + i][j1 + j] : 0;
        }
#endif
        
        //--- store number of nonzero coefficients ---
        best_cnt_nonz  = cnt_nonz;
      
        if (input->rdopt)
        {
          //--- store block cbp ---
          cbp_blk8x8    &= (~(0x33 << (((block>>1)<<3)+((block%2)<<1)))); // delete bits for block
          cbp_blk8x8    |= curr_cbp_blk;
        
          //--- store coefficients ---
          for (k=0; k< 4; k++)
          {
            for (j=0; j< 2; j++)
              for (i=0; i<65; i++)  
                cofACtr[k][j][i] = img->cofAC[block][k][j][i]; // 18->65 for ABT
          }   
          //--- store reconstruction and prediction --- 
          if(!img->residue_transform_flag)
          {
            for (j=j0; j<j0+8; j++)    
            {
#ifdef ADAPTIVE_FD_SD_CODING
              memcpy(&dataTr->quantizer_indices8x8[j][i0],&currMB->quantizer_indices[j][i0],8*sizeof(int));
#endif
              pix_y = img->pix_y + j;
              for (i=i0; i<i0+8; i++)
              {
                pix_x = img->pix_x + i;
                dataTr->rec_mbY8x8[j][i] = enc_picture->imgY[pix_y][pix_x];
                dataTr->mpr8x8[j][i] = img->mpr[j][i];
                if(img->type==SP_SLICE && (!si_frame_indicator))
                  dataTr->lrec[j][i]=lrec[pix_y][pix_x]; // store the coefficients for primary SP slice
              }     
            }
#ifdef ADAPTIVE_FD_SD_CODING
            dataTr->SD_Coding_on_off8x8=currMB->SD_Coding_on_off;
            dataTr->SD_or_FD8x8[j0/8][i0/8]=currMB->SD_or_FD[j0/8][i0/8];
            set_bit(&(dataTr->SD_or_FD_t8x88x8), (j0/8*2+i0/8), (currMB->SD_or_FD_t8x8) & (1<<(j0/8*2+i0/8)));
#endif	     
          }
          else
          {
            for (j=j0; j<j0+8; j++)                
              for (i=i0; i<i0+8; i++)
              {
                dataTr->rec_resG_8x8  [j][i] = rec_resG  [j][i];
                dataTr->resTrans_R_8x8[j][i] = resTrans_R[j][i];
                dataTr->resTrans_B_8x8[j][i] = resTrans_B[j][i];
                dataTr->mprRGB_8x8 [0][j][i] = mprRGB [0][j][i];
                dataTr->mprRGB_8x8 [1][j][i] = mprRGB [1][j][i];
                dataTr->mprRGB_8x8 [2][j][i] = mprRGB [2][j][i];                  
              }     
          }
        }                  
        if (img->AdaptiveRounding)
        {
#ifdef ADAPTIVE_FD_SD_CODING
          if (transform8x8) adjust_adaptive_f_spatial_domain_8x8=img->adjust_adaptive_f_spatial_domain_8x8;
          else              adjust_adaptive_f_spatial_domain_4x4=img->adjust_adaptive_f_spatial_domain_4x4;
#endif	   
          for (j=j0; j<j0+8; j++)
            for (i=i0; i<i0+8; i++)                    
            {
              fadjust     [j][i] = fadjustTransform  [0]   [j][i];
              fadjustCr[0][j][i] = fadjustTransformCr[0][0][j][i];
              fadjustCr[1][j][i] = fadjustTransformCr[0][1][j][i];
            }

          {
              if(gdwFinalEncodeFlag) pAQMSParam = gpBestAdaptiveQuantParam;
              else                   pAQMSParam = gpAdaptiveQuantParam;

              if(transform8x8)
              {
                for (j=j0; j<j0+8; j++)
                  for (i=i0; i<i0+8; i++)                    
                  {
                    fadjust_aqms     [pAQMSParam->dwQMatrixIdx][j][i] = img->fadjust8x8_aqms  [pAQMSParam->dwQMatrixIdx][0]   [j][i];
                    fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][0][j][i] = img->fadjust8x8Cr_aqms[pAQMSParam->dwQMatrixIdx][0][0][j][i];
                    fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][1][j][i] = img->fadjust8x8Cr_aqms[pAQMSParam->dwQMatrixIdx][0][1][j][i];
                  }
              }
              else
              {
                for (j=j0; j<j0+8; j++)
                  for (i=i0; i<i0+8; i++)                    
                  {
                    fadjust_aqms     [pAQMSParam->dwQMatrixIdx][j][i] = img->fadjust4x4_aqms  [pAQMSParam->dwQMatrixIdx][0]   [j][i];
                    fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][0][j][i] = img->fadjust4x4Cr_aqms[pAQMSParam->dwQMatrixIdx][0][0][j][i];
                    fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][1][j][i] = img->fadjust4x4Cr_aqms[pAQMSParam->dwQMatrixIdx][0][1][j][i];
                  }
              }
          }
        }
        //--- store best 8x8 coding state ---
        if (block < 3)
        store_coding_state (cs_b8);
      } // if (rdcost <= min_rdcost)
      
      //--- re-set coding state as it was before coding with current mode was performed ---

      reset_coding_state (cs_cm);
    } // if ((enc_mb.valid[mode] && (transform8x8 == 0 || mode != 0 || (mode == 0 && active_sps->direct_8x8_inference_flag)))
  } // for (min_rdcost=1e30, index=(bslice?0:1); index<6; index++)  
  
#ifdef BEST_NZ_COEFF
  for(j = 0; j <= 1; j++)
  {
    for(i = 0; i <= 1; i++)
      img->nz_coeff[img->current_mb_nr][i1 + i][j1 + j] = best_nz_coeff[i][j];
  }
#endif              
  
  if (!transform8x8)
    dataTr->cost8x8 += min_cost8x8;
  
  //----- set cbp and count of nonzero coefficients ---
  if (best_cnt_nonz)
  {
    cbp8x8       |= (1 << block);
    cnt_nonz_8x8 += best_cnt_nonz;
  }
  
  if (!transform8x8)
  {
    if (block<3)
    {
      //===== re-set reconstructed block =====
      j0   = 8*(block/2);
      i0   = 8*(block%2);
      for (j=j0; j<j0 + 2 * BLOCK_SIZE; j++)
      {
        memcpy(&enc_picture->imgY[img->pix_y + j][img->pix_x], dataTr->rec_mbY8x8[j], 2 * BLOCK_SIZE * sizeof(imgpel));
        if(img->type==SP_SLICE &&(!si_frame_indicator))
          memcpy(&lrec[img->pix_y + j][img->pix_x], dataTr->lrec[j],2*BLOCK_SIZE*sizeof(imgpel)); // reset the coefficients for SP slice
      }
#ifdef ADAPTIVE_FD_SD_CODING
       currMB->SD_Coding_on_off=dataTr->SD_Coding_on_off8x8;
       for (j=j0; j<j0+2* BLOCK_SIZE; j++)
       {
         memcpy(&currMB->quantizer_indices[j][i0],&dataTr->quantizer_indices8x8[j][i0],2* BLOCK_SIZE*sizeof(int));
       }
       currMB->SD_or_FD[j0/8][i0/8]=dataTr->SD_or_FD8x8[j0/8][i0/8];
       set_bit(&(currMB->SD_or_FD_t8x8), (j0/8*2+i0/8), (dataTr->SD_or_FD_t8x88x8) & (1<<(j0/8*2+i0/8)));
#endif
    } // if (block<3)    
  }  
  else
  {  
    //======= save motion data for 8x8 partition for transform size 8x8 ========
    StoreNewMotionVectorsBlock8x8(0, block, dataTr->part8x8mode[block], dataTr->part8x8fwref[block], dataTr->part8x8bwref[block], dataTr->part8x8pdir[block], bslice);
  }  
  //===== set motion vectors and reference frames (prediction) =====
  SetRefAndMotionVectors (block, dataTr->part8x8mode[block], dataTr->part8x8pdir[block], dataTr->part8x8fwref[block], dataTr->part8x8bwref[block]);
  
  //===== set the coding state after current block =====
  //if (transform8x8 == 0 || block < 3)
  if (block < 3)
    reset_coding_state (cs_b8);
  
  if (img->AdaptiveRounding)
  {
    for (j=j0; j<j0+2 * BLOCK_SIZE; j++)
    {
#ifdef ADAPTIVE_FD_SD_CODING
      if      (lumaAdjustIndex==0 &&  transform8x8) img->adjust_adaptive_f_spatial_domain_8x8=adjust_adaptive_f_spatial_domain_8x8;
      else if (lumaAdjustIndex==0 && !transform8x8) img->adjust_adaptive_f_spatial_domain_4x4=adjust_adaptive_f_spatial_domain_4x4;
#endif
      memcpy(&fadjustTransform  [lumaAdjustIndex  ]   [j][i0], &fadjust     [j][i0], 2 * BLOCK_SIZE * sizeof(int));
      memcpy(&fadjustTransformCr[chromaAdjustIndex][0][j][i0], &fadjustCr[0][j][i0], 2 * BLOCK_SIZE * sizeof(int));
      memcpy(&fadjustTransformCr[chromaAdjustIndex][1][j][i0], &fadjustCr[1][j][i0], 2 * BLOCK_SIZE * sizeof(int));
    }

    {
      if(gdwFinalEncodeFlag) pAQMSParam = gpBestAdaptiveQuantParam;
      else                   pAQMSParam = gpAdaptiveQuantParam;

      if(transform8x8)
      {
        for (j=j0; j<j0+2 * BLOCK_SIZE; j++)
          for(i=i0;i<i0+2 * BLOCK_SIZE; i++)
          {
            img->fadjust8x8_aqms  [pAQMSParam->dwQMatrixIdx][lumaAdjustIndex  ]   [j][i] = fadjust_aqms     [pAQMSParam->dwQMatrixIdx][j][i];
            img->fadjust8x8Cr_aqms[pAQMSParam->dwQMatrixIdx][chromaAdjustIndex][0][j][i] = fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][0][j][i];
            img->fadjust8x8Cr_aqms[pAQMSParam->dwQMatrixIdx][chromaAdjustIndex][1][j][i] = fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][1][j][i];
          }
      }
      else
      {
        for (j=j0; j<j0+2 * BLOCK_SIZE; j++)
          for(i=i0;i<i0+2 * BLOCK_SIZE; i++)
          {
            img->fadjust4x4_aqms  [pAQMSParam->dwQMatrixIdx][lumaAdjustIndex  ]   [j][i] = fadjust_aqms     [pAQMSParam->dwQMatrixIdx][j][i];
            img->fadjust4x4Cr_aqms[pAQMSParam->dwQMatrixIdx][chromaAdjustIndex][0][j][i] = fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][0][j][i];
            img->fadjust4x4Cr_aqms[pAQMSParam->dwQMatrixIdx][chromaAdjustIndex][1][j][i] = fadjustCr_aqms[pAQMSParam->dwQMatrixIdx][1][j][i];
          }
      }
    }
  }
}

/*!
*************************************************************************************
* \brief
*    Quantization RD decision process
*************************************************************************************
*/
void ComputeQuantizationRDcost(int mode, 
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short best_ctr16x16,
                          short best_cbp,
                          short *best_mode,
                          short *best_matrix,
                          short *best_qp)
{
// perform Quantization rate - distortion optimization

  gpAdaptiveQuantParam->dwQMatrixIdx = 0;
#ifdef ADAPTIVE_QP
  gpAdaptiveQuantParam->dwCurrQP = img->qp;
#endif

  {
#ifdef ADAPTIVE_QP
    if(input->UseAdaptiveQuantParam && best_cbp!=0)
    {
      DecideBestQPForMB(mode, currMB, enc_mb, min_rdcost, min_rate, i16mode, bslice, inter_skip, best_ctr16x16, best_mode, best_qp);
      img->qp = currMB->qp = gpAdaptiveQuantParam->dwCurrQP = gpBestAdaptiveQuantParam->dwCurrQP = *best_qp;
      set_chroma_qp(currMB);
      currMB->delta_qp = (currMB->qp - currMB->prev_qp);
    }
#endif

#ifdef ADAPTIVE_QUANTIZATION
    if(img->slice_adaptive_quant_matrix_flag && best_cbp!=0)
    {
      DecideBestMatrixForMB(mode, currMB, enc_mb, min_rdcost, min_rate, i16mode, bslice, inter_skip, best_ctr16x16, best_mode, best_matrix);
      gpBestAdaptiveQuantParam->dwQMatrixIdx = gpAdaptiveQuantParam->dwQMatrixIdx = *best_matrix;
    }
#endif
  }
  
};

/*!
*************************************************************************************
* \brief
*    Final encoding for the best quantization parameter for the best mode
*************************************************************************************
*/
void ComputeBestQuantizationRDCost( int best_mode, 
                                        Macroblock *currMB, 
                                        RD_PARAMS enc_mb, 
                                        double *min_rdcost, 
                                        double *min_rate, 
                                        int i16mode, 
                                        short bslice, 
                                        int   rd_qp,
                                        short *best_bi_pred_me,
                                        short *inter_skip,
                                        short best_ctr16x16,
                                        short best_cbp,
                                        short *best_matidx,
                                        short *best_qp)
{
  //  short dwsummy=0;
  double min_rdcost_tmp = 1e30;
  double min_rate_tmp   = 1e30;
  currMB->c_ipred_mode = best_c_imode;

  if(currMB->c_ipred_mode > 3 || currMB->c_ipred_mode < 0)
  {
    printf("illegal chroma pred mode!!\n"); // for error evasion
    assert(currMB->c_ipred_mode < 4);
  }
  if(input->BiPredMotionEstimation)
    img->bi_pred_me[best_mode] = best_bi_pred_me[best_mode];

  // adaptive quantization parameter setting for syntax
  if(best_mode == I16MB || best_mode == IPCM) *best_matidx = 0;
#ifdef ADAPTIVE_QUANTIZATION
  if(img->slice_adaptive_quant_matrix_flag) currMB->aqms_idx = gpAdaptiveQuantParam->dwQMatrixIdx = gpBestAdaptiveQuantParam->dwQMatrixIdx = *best_matidx;
  else  currMB->aqms_idx = gpAdaptiveQuantParam->dwQMatrixIdx = gpBestAdaptiveQuantParam->dwQMatrixIdx = *best_matidx = 0;
#endif
#ifdef ADAPTIVE_QP
  if(input->UseAdaptiveQuantParam){
    currMB->qp = img->qp = gpAdaptiveQuantParam->dwCurrQP = gpBestAdaptiveQuantParam->dwCurrQP = *best_qp;
    set_chroma_qp(currMB);
    currMB->delta_qp = (currMB->qp - currMB->prev_qp);
  }
  else
  {
    currMB->qp = img->qp = gpAdaptiveQuantParam->dwCurrQP = gpBestAdaptiveQuantParam->dwCurrQP = *best_qp = rd_qp;
    set_chroma_qp(currMB);
    currMB->delta_qp = (currMB->qp - currMB->prev_qp);
  }
#endif

#ifdef ADAPTIVE_QUANTIZATION
  // coeff consideration for weighted quantization matrix
  ComputeCoeffRDCost(best_mode, currMB, enc_mb, min_rdcost, min_rate, i16mode, bslice, inter_skip, best_ctr16x16, best_cbp);
  //reset_coding_state (cs_cm);
#endif

  gdwFinalEncodeFlag = 1;
//******************* re-set the P or B parameter ***********************************************************************************************************************
  PreSetBestCodingParameters(best_mode, currMB, enc_mb, bslice, inter_skip, best_ctr16x16);
//******************* encdoe best aqms_idx **************************************************************************************************************************

  compute_mode_RD_cost_Final(best_mode, currMB, enc_mb, &min_rdcost_tmp, &min_rate_tmp, i16mode, bslice, inter_skip);
#ifdef ADAPTIVE_QUANTIZATION
  if( gdwBestCbp == 0 || best_mode == I16MB || (img->type==P_SLICE && currMB->skip_flag && input->symbol_mode == CABAC) )
    currMB->aqms_idx = 0;
#endif

  input->disthres = gdwDefaultdisthres;
}

/*!
 ************************************************************************
 * \brief
 *    Initialize quantization parameters
 *
 ************************************************************************
 */
void InitAdaptiveQuantizationParam (void)
{
  gdwFinalEncodeFlag = 0;
  gdwBestCbp = 0;

  gpAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;
  gpAdaptiveQuantParam->dwQMatrixIdx = 0;
  gpAdaptiveQuantParam->dwCoeffIdx = 0;
  gpBestAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;
  gpBestAdaptiveQuantParam->dwQMatrixIdx = 0;
  gpBestAdaptiveQuantParam->dwCoeffIdx = 0;

#ifdef ADAPTIVE_QUANTIZATION
  gpAdaptiveQuantParam->dwAdaptiveQMOptFlag = img->AdaptiveQMOptFlag;
  gpAdaptiveQuantParam->dwAdaptiveCoeffFlag    = img->AdaptiveCoeffFlag;
  gpAdaptiveQuantParam->dwUseQMatrixFlag = img->slice_adaptive_quant_matrix_flag;
  gpBestAdaptiveQuantParam->dwAdaptiveQMOptFlag = img->AdaptiveQMOptFlag;
  gpBestAdaptiveQuantParam->dwAdaptiveCoeffFlag    = img->AdaptiveCoeffFlag;
  gpBestAdaptiveQuantParam->dwUseQMatrixFlag = img->slice_adaptive_quant_matrix_flag;

  gpModelingMatrixType->gdwMatrixGradient   = 32;
  gpModelingMatrixType->gdwMatrixDCgain     = 16;
#else
  gpAdaptiveQuantParam->dwAdaptiveQMOptFlag = 0;
  gpAdaptiveQuantParam->dwAdaptiveCoeffFlag = 0;
  gpAdaptiveQuantParam->dwUseQMatrixFlag = 0;
  gpBestAdaptiveQuantParam->dwAdaptiveQMOptFlag = 0;
  gpBestAdaptiveQuantParam->dwAdaptiveCoeffFlag = 0;
  gpBestAdaptiveQuantParam->dwUseQMatrixFlag = 0;
#endif

#ifdef ADAPTIVE_QP
  gpAdaptiveQuantParam->dwCurrQP = img->qp;
  gpBestAdaptiveQuantParam->dwCurrQP = img->qp;
  gpAdaptiveQuantParam->dwUseAdaptiveQuantParam = input->UseAdaptiveQuantParam;
  gpBestAdaptiveQuantParam->dwUseAdaptiveQuantParam = input->UseAdaptiveQuantParam;
#endif
}
/*!
*************************************************************************************
* \brief
*    computing the same lambda for different QP values
*************************************************************************************
*/
void CalculateLambdaForMatrix(RD_PARAMS   *enc_mb)
{
  short       bslice      = (img->type==B_SLICE);

  // Anchor QP setting
#ifdef ADAPTIVE_QUANTIZATION
#ifdef ADAPTIVE_QP
  if(input->UseAdaptiveQuantMatrix || input->UseAdaptiveQuantParam)
#else
  if(input->UseAdaptiveQuantMatrix)
#endif
#else
#ifdef ADAPTIVE_QP
  if(input->UseAdaptiveQuantParam)
#endif
#endif
  {
    if (bslice && img->nal_reference_idc)
    {
      enc_mb->lambda_md = img->lambda_md[5][gdwBaseQp];
      enc_mb->lambda_me = img->lambda_me[5][gdwBaseQp];
      enc_mb->lambda_mf = img->lambda_mf[5][gdwBaseQp];
    }
    else
    {
      enc_mb->lambda_md = img->lambda_md[img->type][gdwBaseQp];
      enc_mb->lambda_me = img->lambda_me[img->type][gdwBaseQp];
      enc_mb->lambda_mf = img->lambda_mf[img->type][gdwBaseQp];
    }
  }
}
#endif

#ifdef ADAPTIVE_QUANTIZATION

//------------------------- for image.c
/*!
 ***********************************************************************
 * \brief
 *   adaptive quantization matrix selection
 ***********************************************************************
 */
int rdPictureCodingAQMS(int AQMS_Skip_flag, sDefaultInputParam_t *defInputParam, int *mincost_flg, int rd_qp)
{
  int ModelingMatrix_Skip_Flag = 0;

#ifdef ADAPTIVE_QP
  if(input->UseAdaptiveQuantMatrix) input->UseAdaptiveQuantParam = 0;
#endif
  ReGenerateParameterSets(active_pps->pic_parameter_set_id);

  // tentative encoding for frame

  if(!AQMS_Skip_flag)
    ModelingMatrix_Skip_Flag = rdPictureCodingForModelingMatrixDesign(rd_qp);

  // return default input parameter
  LoadInputParam(defInputParam);
  if(AQMS_Skip_flag || ModelingMatrix_Skip_Flag)
  {
    img->slice_adaptive_quant_matrix_flag = 0;
    ResetModelingParam();
  }

  // Final encode
  ResetFrameMemory();
  FramePcitureForAQMS (frame_pic_aqms,1,BestScaleIdx);

  *mincost_flg = ChooseBestPicture(rd_qp);

  if(*mincost_flg)
    ResetModelingParam();

#ifdef ADAPTIVE_QP
  input->UseAdaptiveQuantParam = useQPvalueFlag;
#endif
#ifdef MV_COMPETITION // MVC bug fix by AT
  if(*mincost_flg && input->mv_competition)
  {
    img->slice_adaptive_quant_matrix_flag = 0;
    if(enc_frame_picture)
    {
      free_storable_picture (enc_frame_picture);
      enc_frame_picture = NULL;
    }
    if (frame_pic_1)
    {
      free_slice_list(frame_pic_1);
    }
    frame_picture (frame_pic_1,img->rd_pass);
  }
#endif

  SkipRatio[(img->type)+(img->type==B_SLICE? (img->nal_reference_idc<<2) : 0)] = ((double)dwSkipMbCountForSlice[(img->type)+(img->type==B_SLICE? (img->nal_reference_idc<<2) : 0)]/(double)dwSkipEnableMbCountForSlice[(img->type)+(img->type==B_SLICE? (img->nal_reference_idc<<2) : 0)]*100.0);

  return 0;
}

/*!
 ***********************************************************************
 * \brief
 *   calculating the initial modeling parameters
 ***********************************************************************
 */
void CalculateInitinalModelingParamters(int param_idx, int *ParamA, int *ParamB)
{
                                   //P-slice B-slice I-slice SP      SI
  const int cDefaultParamA[5][2] = {{22,13},{19,12},{24,14},{19,12},{19,12}};
  const int cDefaultParamB[5][2] = {{52,19},{64,20},{48,18},{64,20},{64,20}};

  if(img->type==I_SLICE)
  {
    *ParamA = cDefaultParamA[1][param_idx];
    *ParamB = cDefaultParamB[1][param_idx];
  }
  else
  {
    if(param_idx==0)
    {
      *ParamA = Clip3(0, 112, (SliceMatrixParam0[img->type]-16) );
      *ParamB = Clip3(0, 112, (SliceMatrixParam1[img->type]+32) );
    }
    else
    {
      *ParamA = Clip3(10, 22, (SliceMatrixParam0[img->type]-3) );
      *ParamB = Clip3(10, 22, (SliceMatrixParam1[img->type]+3) );
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *   the optimal modeling quantization matrix calculation routines
 ***********************************************************************
 */
int rdPictureCodingForModelingMatrixDesign(int rd_qp)
{
  int S1 = 0, S2 = 0, R1, R2, D1, D2;
  int Gradient=32, DCgain=16;
  double lambda;
  sModelingMatrixType_t ModelingMatrixType;

  BestScaleIdx=0;

  // initialize matrix parameter
  SetModelingMatrixParam(&ModelingMatrixType, 0, 32, 16);

  CalculateInitinalModelingParamters(1, &S1, &S2);
  // set first DCgain
  ModelingMatrixType.gdwMatrixDCgain = S1;
  PictureCodingDecisionForMatrix(&ModelingMatrixType, &R1, &D1, &lambda, rd_qp);
  // set second DCgain
  ModelingMatrixType.gdwMatrixDCgain = S2;
  PictureCodingDecisionForMatrix(&ModelingMatrixType, &R2, &D2, &lambda, rd_qp);
  // estimation the best DCgain parameter
  DCgain = CalculateBestMatrixParam(1, S1, S2, R1, R2, D1, D2, lambda, rd_qp);
  ModelingMatrixType.gdwMatrixDCgain = DCgain;

  CalculateInitinalModelingParamters(0, &S1, &S2);
  // set first gradient
  ModelingMatrixType.gdwMatrixGradient = S1;
  PictureCodingDecisionForMatrix(&ModelingMatrixType, &R1, &D1, &lambda, rd_qp);
  // set second gradient
  ModelingMatrixType.gdwMatrixGradient = S2;
  PictureCodingDecisionForMatrix(&ModelingMatrixType, &R2, &D2, &lambda, rd_qp);
  // estimation the best gradient parameter
  Gradient = CalculateBestMatrixParam(0, S1, S2, R1, R2, D1, D2, lambda, rd_qp);
  ModelingMatrixType.gdwMatrixGradient = Gradient;

  SetModelingMatrixParam(&ModelingMatrixType, 0, Gradient, DCgain);
  VerifyModelingParam(&ModelingMatrixType);
  SetActiveMatrixFromParam(0, &ModelingMatrixType);
  if(img->AdaptivePicRDOFlag)
  {
    SetModelingMatrixParam(&ModelingMatrixType, 0, 32, 16);
    SetActiveMatrixFromParam(1, &ModelingMatrixType);
  }

  SliceMatrixParam0[img->type]=ModelingMatrixType.gdwMatrixGradient;
  SliceMatrixParam1[img->type]=ModelingMatrixType.gdwMatrixDCgain;

  // same matrix check
  if(CheckSameModelingParam()) return 1;

  img->write_macroblock = 0;

  {
    enc_picture = enc_frame_picture_aqms;
    frame_pic   = frame_pic_aqms;
    active_pps  = PicParSet[3];
  }

  return 0;
}

/*!
 ***********************************************************************
 * \brief
 *   create quantization matrix for the modeling parameters
 ***********************************************************************
 */
void SetActiveMatrixFromParam(int Idx, sModelingMatrixType_t *ModelingMatrixType)
{
  img->slice_scaling_model_param0[Idx]   = ModelingMatrixType->gdwMatrixGradient;
  img->slice_scaling_model_param1[Idx]   = ModelingMatrixType->gdwMatrixDCgain;

  // set scaling parameters
  gpModelingMatrixType->gdwMatrixGradient   = ModelingMatrixType->gdwMatrixGradient;
  gpModelingMatrixType->gdwMatrixDCgain     = ModelingMatrixType->gdwMatrixDCgain;

  // create  quantization matrices 
  SetAutoScalingListData(Idx);
}
/*!
 ***********************************************************************
 * \brief
 *   tentative encoding process for computing the best modeling parameters
 ***********************************************************************
 */
void PictureCodingDecisionForMatrix(sModelingMatrixType_t *ModelingMatrixType, int *Rate, int *Dist, double *lambda, int rd_qp)
{
  double lambda_picture;
  int spframe = (img->type == SP_SLICE);
  int bframe = (img->type == B_SLICE);

  active_pps  = PicParSet[3];
  enc_picture = enc_frame_picture_aqms;
  frame_pic   = frame_pic_aqms;

  SetActiveMatrixFromParam(0, ModelingMatrixType);

  ResetFrameMemory();
  FramePcitureForAQMS (frame_pic_aqms,1,0);

  if (input->successive_Bframe)
    lambda_picture = 0.68 * pow (2, (rd_qp - SHIFT_QP) / 3.0) * (bframe || spframe ? 2 : 1);    
  else
    lambda_picture = 0.68 * pow (2, (rd_qp - SHIFT_QP) / 3.0);

  *lambda = lambda_picture;

  *Dist = (int)( (float)(frame_pic_aqms->distortion_y + frame_pic_aqms->distortion_u + frame_pic_aqms->distortion_v) );  
  *Rate = frame_pic_aqms->bits_per_picture;
}
/*!
 ***********************************************************************
 * \brief
 *   calculating the best modeling parameter
 ***********************************************************************
 */
int CalculateBestMatrixParam(int param_idx, int S1, int S2, int R1, int R2, int D1, int D2, double lambda, int rd_qp)
{
  const double cMIN_VALUE = 1.0e-040;
  double EstR=777.0, EstParam = 16.0;
  double a=0.0, b=0.0, A=0.0, B=0.0, numer=0.0, denom=0.0;
  int    skip_flag=0, outlier_flag=0, error_flag=0;

  //assert(R1>0 && R2>0 && D1>0 && D2>0);
  if (R1<0 || R2<0 || D1<0 || D2<0)
    error_flag = 1;

  if (R1 > PIC_SKIP_RATE && R2 > PIC_SKIP_RATE && !error_flag)
  {
    numer = (log((double)R2)-log((double)R1));
    denom = (double)(S1 - S2);
    if(fabs(denom)>cMIN_VALUE)
    {
      B = -(numer)/denom;
      A = ((double)S1*log((double)R2)-(double)S2*log((double)R1))/denom;
    }
    else
      skip_flag = 1;
  
    denom = (log((double)D2)-log((double)D1));
    if(fabs(denom)>cMIN_VALUE)
    {
      a = (numer/denom);
      numer = (log((double)R1)*log((double)D2)-log((double)R2)*log((double)D1));
      b = exp(numer/denom);
      if(fabs(a)>cMIN_VALUE)
        denom = (-a*lambda*pow(b, (1.0/a)) );
      else
        skip_flag = 1;

      if(denom>=cMIN_VALUE && fabs(1.0-a)>cMIN_VALUE)
        EstR = (pow(denom, a/(1.0-a)) );
      else
        skip_flag = 1;
    }
    else
      skip_flag = 1;

    numer = (log(EstR) - A);
    if(fabs(B)>cMIN_VALUE)
      EstParam = numer/B;
    else
      skip_flag = 1;

  }
  else
    skip_flag = 1;

  outlier_flag = CheckModelingParam(param_idx, EstParam);

  if(skip_flag || outlier_flag)
  {
    if(param_idx==0)
      EstParam = 32.0;
    else
      EstParam = 16.0;
  }
  return (int)(EstParam+0.5);

}
/*!
 ***********************************************************************
 * \brief
 *   set modeling parameter
 ***********************************************************************
 */
void SetModelingMatrixParam(sModelingMatrixType_t *ModelingMatrixType, int type, int gradient,  int DCgain)
{
  ModelingMatrixType->gdwMatrixGradient   = gradient;
  ModelingMatrixType->gdwMatrixDCgain     = DCgain;
}
/*!
 ***********************************************************************
 * \brief
 *   reset frame memory
 ***********************************************************************
 */
void ResetFrameMemory()
{
  if(enc_frame_picture_aqms)
  {
    free_storable_picture (enc_frame_picture_aqms);
    enc_frame_picture_aqms = NULL;
  }
  if (frame_pic_aqms)
  {
    free_slice_list(frame_pic_aqms);
  }
}
/*!
 ***********************************************************************
 * \brief
 *   checking modeling parameters to avoid the misplaced values
 ***********************************************************************
 */
int CheckModelingParam(int param_idx, double EstParam)
{
  int outlier_flag=0;

  if(param_idx==0)
  {
    if(0.0 > EstParam || EstParam > 112.0)
      outlier_flag = 1;
  }
  else
  {
    if(10.0 > EstParam || EstParam > 22.0)
      outlier_flag = 1;
  }

  return outlier_flag;
}
/*!
 ***********************************************************************
 * \brief
 *   verify the modeling parameters to avoid the outliar
 ***********************************************************************
 */
void VerifyModelingParam(sModelingMatrixType_t *ModelingMatrixType)
{
  if (img->type == P_SLICE)
  {
    // Outlier correction
  }
  else if(img->type == I_SLICE)
  {
    // Outlier correction
    if(ModelingMatrixType->gdwMatrixDCgain > 15 || ModelingMatrixType->gdwMatrixDCgain < 12)
      SetModelingMatrixParam(ModelingMatrixType, 0, 32, 15);
  }
  else
  {
    // Outlier correction
    if(ModelingMatrixType->gdwMatrixGradient < 32 && ModelingMatrixType->gdwMatrixDCgain < 16)
      // initialize matrix parameter
      SetModelingMatrixParam(ModelingMatrixType, 0, 32, ModelingMatrixType->gdwMatrixDCgain);
  }
}
/*!
 ***********************************************************************
 * \brief
 *   checking the same quantization matrix parameters
 ***********************************************************************
 */
int CheckSameModelingParam()
{
  int valid_flag = 0;

  valid_flag += ( img->slice_scaling_model_param0   [0] != img->slice_scaling_model_param0[1]);
  valid_flag += ( img->slice_scaling_model_param1   [0] != img->slice_scaling_model_param1[1]);

  if(valid_flag)
    return 0;
  else
    return 1;
}
/*!
 ***********************************************************************
 * \brief
 *   storing the input parameters (for speed up)
 ***********************************************************************
 */
void SaveInputParam(int AQMS_Skip_flag, sDefaultInputParam_t *defInputParam)
{
  defInputParam->default_AQMS_flag    =img->slice_adaptive_quant_matrix_flag;
  defInputParam->default_QMOpt_flag   =img->AdaptiveQMOptFlag;
  defInputParam->default_Coeff_flag   =img->AdaptiveCoeffFlag;
  defInputParam->default_me_modeflag  =input->FMEnable;
  defInputParam->default_bipred_me_flag = input->BiPredMotionEstimation;
  defInputParam->default_BiPredMESearchRange = input->BiPredMESearchRange;

  // storing the additional command options for speed up encoding.
  // Please add the new command option if you make.
#ifdef ADAPTIVE_FD_SD_CODING
  defInputParam->default_APEC_in_FD_and_SD = input->APEC_in_FD_and_SD;
#endif
#ifdef MV_COMPETITION
  defInputParam->default_MVCompetition = input->mv_competition;
#endif

  // for fast encoding
  img->slice_adaptive_quant_matrix_flag   = 0;
  input->FMEnable            = 3;
  input->BiPredMESearchRange = 8;
  img->AdaptiveQMOptFlag	   = 0;
  img->AdaptiveCoeffFlag     = 0;

  // swtiching off the additional command options for speed up encoding.
  // Please add the new command option if you make.
#ifdef ADAPTIVE_FD_SD_CODING
  input->APEC_in_FD_and_SD = 0;
#endif
#ifdef MV_COMPETITION
  input->mv_competition = 0;
#endif

}
/*!
 ***********************************************************************
 * \brief
 *   loading the input parameters (for speed up)
 ***********************************************************************
 */
void LoadInputParam(sDefaultInputParam_t *defInputParam)
{
  img->slice_adaptive_quant_matrix_flag   = defInputParam->default_AQMS_flag;
  img->AdaptiveQMOptFlag	    = defInputParam->default_QMOpt_flag;
  img->AdaptiveCoeffFlag      = defInputParam->default_Coeff_flag;
  input->FMEnable             = defInputParam->default_me_modeflag;
  input->BiPredMESearchRange  = defInputParam->default_BiPredMESearchRange;

  // loading the additional command option for speed up encoding.
  // Please add the new command option if you make.
#ifdef ADAPTIVE_FD_SD_CODING
  input->APEC_in_FD_and_SD    = defInputParam->default_APEC_in_FD_and_SD;
#endif
#ifdef MV_COMPETITION
  input->mv_competition       = defInputParam->default_MVCompetition;
#endif

}
/*!
 ***********************************************************************
 * \brief
 *   tentative encoding to decide the modeling parameters
 ***********************************************************************
 */
void FramePcitureForAQMS (Picture *frame, int rd_pass, int iScaleIdx)
{
 
  img->structure = FRAME;
  img->PicSizeInMbs = img->FrameSizeInMbs;

  if (rd_pass == 0)
  {
    enc_frame_picture  = alloc_storable_picture (img->structure, img->width, img->height, img->width_cr, img->height_cr);
    img->ThisPOC=enc_frame_picture->poc=img->framepoc;
    enc_frame_picture->top_poc    = img->toppoc;
    enc_frame_picture->bottom_poc = img->bottompoc;
    
    enc_frame_picture->frame_poc = img->framepoc;
    
    enc_frame_picture->pic_num = img->frame_num;
    enc_frame_picture->frame_num = img->frame_num;
    enc_frame_picture->coded_frame = 1;
    
    enc_frame_picture->MbaffFrameFlag = img->MbaffFrameFlag = (input->MbInterlace != FRAME_CODING);
    
    enc_picture=enc_frame_picture;
    copy_params();
  }
  else
  {
    {
      enc_frame_picture_aqms  = alloc_storable_picture (img->structure, img->width, img->height, img->width_cr, img->height_cr);
      img->ThisPOC=enc_frame_picture_aqms->poc=img->framepoc;
      enc_frame_picture_aqms->top_poc    = img->toppoc;
      enc_frame_picture_aqms->bottom_poc = img->bottompoc;
    
      enc_frame_picture_aqms->frame_poc = img->framepoc;
    
      enc_frame_picture_aqms->pic_num = img->frame_num;
      enc_frame_picture_aqms->frame_num = img->frame_num;
      enc_frame_picture_aqms->coded_frame = 1;
    
      enc_frame_picture_aqms->MbaffFrameFlag = img->MbaffFrameFlag = (input->MbInterlace != FRAME_CODING);
    
      enc_picture=enc_frame_picture_aqms;
      copy_params();
    }
  }

  stats->em_prev_bits_frm = 0;
  stats->em_prev_bits = &stats->em_prev_bits_frm;

  img->fld_flag = 0;

#ifdef ADAPTIVE_FD_SD_CODING
  img->APEC_in_FD_and_SD = input->APEC_in_FD_and_SD;
#endif

  code_a_picture(frame);

  frame->bits_per_picture = 8 * ((((img->currentSlice)->partArr[0]).bitstream)->byte_pos);
  
  if (img->structure==FRAME)
  {
    find_distortion ();      
    frame->distortion_y = snr->snr_y;
    frame->distortion_u = snr->snr_u;
    frame->distortion_v = snr->snr_v;
  }
}
/*!
 ***********************************************************************
 * \brief
 *   checking RD cost for previous flat quantization matrix
 ***********************************************************************
 */
int ChooseBestPicture(int rd_qp)
{
  int mincost_flg=0;
  img->write_macroblock = 0;

  if(input->RDPictureDecision && img->rd_pass == 2)
  {
    mincost_flg=picture_coding_decision_for_AQMS(frame_pic_aqms, frame_pic_3, rd_qp);
    if(mincost_flg)
    {
      active_pps = PicParSet[active_pps->pic_parameter_set_id];
      enc_picture = enc_frame_picture3;
      frame_pic   = frame_pic_3;
    }
  }
  else if(input->RDPictureDecision && img->rd_pass == 1)
  {
    mincost_flg=picture_coding_decision_for_AQMS(frame_pic_aqms, frame_pic_2, rd_qp);
    if(mincost_flg)
    {
      active_pps = PicParSet[active_pps->pic_parameter_set_id];
      enc_picture = enc_frame_picture2;
      frame_pic   = frame_pic_2;
    }
  }
  else
  {
    mincost_flg=picture_coding_decision_for_AQMS(frame_pic_aqms, frame_pic_1, rd_qp);
    if(mincost_flg)
    {
      active_pps = PicParSet[active_pps->pic_parameter_set_id];
      enc_picture = enc_frame_picture;
      frame_pic   = frame_pic_1;
    }
  }

  return mincost_flg;

}
/*!
 ***********************************************************************
 * \brief
 *   reset the modeling quantization parameters
 ***********************************************************************
 */
void ResetModelingParam (void)
{
  sModelingMatrixType_t ModelingMatrixType;
  SetModelingMatrixParam(&ModelingMatrixType, 0, 32, 16);
  SetActiveMatrixFromParam(0, &ModelingMatrixType);
  SetActiveMatrixFromParam(1, &ModelingMatrixType);
}

//------------------------- for mode_decision.c

/*!
*************************************************************************************
* \brief
*    Modeling Quantization Matrix Cntrol for a macroblock
*************************************************************************************
*/
void DecideBestMatrixForMB(int mode, 
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short best_ctr16x16,
                          short *best_mode,
                          short *best_matrix)
{
  const unsigned int cMaxMatrix = 2;
  unsigned int dwMatIdx;
  double cost_matrix = *min_rdcost;

  gpAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;
  gpAdaptiveQuantParam->dwQMatrixIdx = 0;

  for(dwMatIdx=0;dwMatIdx<cMaxMatrix;dwMatIdx++)
  {
    if(mode == I16MB) continue;

    // reset the optimal parameters
    gpAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;
    gpAdaptiveQuantParam->dwQMatrixIdx = dwMatIdx;
    currMB->aqms_idx = dwMatIdx;
#ifdef ADAPTIVE_QP
    currMB->qp = img->qp = gpBestAdaptiveQuantParam->dwCurrQP;
    set_chroma_qp(currMB);
    currMB->delta_qp = (currMB->qp - currMB->prev_qp);
#endif
//******************* re-set the P or B parameter ***********************************************************************************************************************
    PreSetBestCodingParameters(mode, currMB, enc_mb, bslice, inter_skip, best_ctr16x16);
//******************* decide best aqms_idx ***************************************************************************************************************************

    //--- transform size ---           
    currMB->luma_transform_size_8x8_flag = input->Transform8x8Mode==2
      ?  (mode >= 1 && mode <= 3)
      || (mode == 0 && bslice && active_sps->direct_8x8_inference_flag)
      || ((mode == P8x8) && (enc_mb.valid[4]))
      :  0;

    SetModesAndRefframeForBlocks (mode);
    // Encode with coefficients
    img->NoResidueDirect = 0;

    {
      while(1)
      {
        if(RDCost_for_macroblocks (enc_mb.lambda_md, mode, &cost_matrix, min_rate, i16mode))
        {
          if(cost_matrix < *min_rdcost)
          {
            *best_matrix = dwMatIdx;
            *min_rdcost = cost_matrix;
            store_macroblock_parameters (mode);
            gdwBestCbp = currMB->cbp;
          }
        }

        // Go through transform modes. 
        // Note that if currMB->cbp is 0 one could choose to skip 8x8 mode
        // although this could be due to deadzoning decisions.
        //if (input->Transform8x8Mode==1 && currMB->cbp!=0) 
        if (input->Transform8x8Mode==1)
        {
          //=========== try mb_types 1,2,3 with 8x8 transform ===========
          if ((mode >= 1 && mode <= 3) && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try DIRECT-MODE with 8x8 transform ===========
          else if (mode == 0 && bslice && active_sps->direct_8x8_inference_flag && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try mb_type P8x8 for mode 4 with 4x4/8x8 transform ===========
          else if ((mode == P8x8) && (enc_mb.valid[4]) && (currMB->luma_transform_size_8x8_flag == 0))
          {
            currMB->luma_transform_size_8x8_flag = 1; //check 8x8 partition for transform size 8x8
            continue;
          }
          else
          {
            currMB->luma_transform_size_8x8_flag = 0;
            break;
          }
        }
        else
          break;
      }
    }

  }
}
/*!
*************************************************************************************
* \brief
*    Quantization Cntrol for a macroblock
*************************************************************************************
*/
void DecideBestRdoForMB(int mode, 
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short best_ctr16x16,
                          short *best_rdo)
{
  unsigned int cMaxRDOidx = 16;
  unsigned int dwRDOIdx;
  double cost_offset = *min_rdcost;

  gpAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;

  for(dwRDOIdx=0;dwRDOIdx<cMaxRDOidx;dwRDOIdx++)
  {
    // reset the optimal parameters
    gpAdaptiveQuantParam->dwAdaptiveQPOIdx = dwRDOIdx;
    gpAdaptiveQuantParam->dwQMatrixIdx = currMB->aqms_idx = gpBestAdaptiveQuantParam->dwQMatrixIdx;
#ifdef ADAPTIVE_QP
    currMB->qp = img->qp = gpBestAdaptiveQuantParam->dwCurrQP;
    set_chroma_qp(currMB);
    currMB->delta_qp = (currMB->qp - currMB->prev_qp);
#endif
//******************* re-set the P or B parameter ***********************************************************************************************************************
    PreSetBestCodingParameters(mode, currMB, enc_mb, bslice, inter_skip, best_ctr16x16);
//******************* decide best aqms_idx ***************************************************************************************************************************

    //--- transform size ---           
    currMB->luma_transform_size_8x8_flag = input->Transform8x8Mode==2
      ?  (mode >= 1 && mode <= 3)
      || (mode == 0 && bslice && active_sps->direct_8x8_inference_flag)
      || ((mode == P8x8) && (enc_mb.valid[4]))
      :  0;

    SetModesAndRefframeForBlocks (mode);
    // Encode with coefficients
    img->NoResidueDirect = 0;

    {
      while(1)
      {
        if(RDCost_for_macroblocks (enc_mb.lambda_md, mode, &cost_offset, min_rate, i16mode))
        {
          if(cost_offset < *min_rdcost)
          {
            *best_rdo = dwRDOIdx;
            *min_rdcost = cost_offset;
            store_macroblock_parameters (mode);
            gdwBestCbp = currMB->cbp;
          }
        }

        // Go through transform modes. 
        // Note that if currMB->cbp is 0 one could choose to skip 8x8 mode
        // although this could be due to deadzoning decisions.
        //if (input->Transform8x8Mode==1 && currMB->cbp!=0) 
        if (input->Transform8x8Mode==1)
        {
          //=========== try mb_types 1,2,3 with 8x8 transform ===========
          if ((mode >= 1 && mode <= 3) && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try DIRECT-MODE with 8x8 transform ===========
          else if (mode == 0 && bslice && active_sps->direct_8x8_inference_flag && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try mb_type P8x8 for mode 4 with 4x4/8x8 transform ===========
          else if ((mode == P8x8) && (enc_mb.valid[4]) && (currMB->luma_transform_size_8x8_flag == 0))
          {
            currMB->luma_transform_size_8x8_flag = 1; //check 8x8 partition for transform size 8x8
            continue;
          }
          else
          {
            currMB->luma_transform_size_8x8_flag = 0;
            break;
          }
        }
        else
          break;
      }
    }

  }
}
/*!
*************************************************************************************
* \brief
*    Quantization Coeff Control for a macroblock
*************************************************************************************
*/
void DecideBestCoeffForMB(int mode, 
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short best_ctr16x16,
                          short *best_coeff)
{
  const unsigned int cMaxCoeffidx = 2;
  unsigned int dwCoeffIdx = 1;
  double cost_coeff = *min_rdcost;

  for(dwCoeffIdx=0;dwCoeffIdx<cMaxCoeffidx;dwCoeffIdx++)
  {
    if(gpBestAdaptiveQuantParam->dwQMatrixIdx==1) continue;

    if(!dwCoeffIdx)
      input->disthres = 0;
    else
      input->disthres = 1;

    // reset the optimal parameters
    gpAdaptiveQuantParam->dwAdaptiveQPOIdx = gpBestAdaptiveQuantParam->dwAdaptiveQPOIdx;
    gpAdaptiveQuantParam->dwQMatrixIdx = currMB->aqms_idx = gpBestAdaptiveQuantParam->dwQMatrixIdx;
#ifdef ADAPTIVE_QP
    currMB->qp = img->qp = gpBestAdaptiveQuantParam->dwCurrQP;
    set_chroma_qp(currMB);
    currMB->delta_qp = (currMB->qp - currMB->prev_qp);
#endif
//******************* re-set the P or B parameter ***********************************************************************************************************************
    PreSetBestCodingParameters(mode, currMB, enc_mb, bslice, inter_skip, best_ctr16x16);
//******************* decide best aqms_idx ***************************************************************************************************************************

    //--- transform size ---           
    currMB->luma_transform_size_8x8_flag = input->Transform8x8Mode==2
      ?  (mode >= 1 && mode <= 3)
      || (mode == 0 && bslice && active_sps->direct_8x8_inference_flag)
      || ((mode == P8x8) && (enc_mb.valid[4]))
      :  0;

    SetModesAndRefframeForBlocks (mode);
    // Encode with coefficients
    img->NoResidueDirect = 0;

    {
      while(1)
      {
        if(RDCost_for_macroblocks (enc_mb.lambda_md, mode, &cost_coeff, min_rate, i16mode))
        {
          if(cost_coeff < *min_rdcost)
          {
            *best_coeff = dwCoeffIdx;
            *min_rdcost = cost_coeff;
            store_macroblock_parameters (mode);
            gdwBestCbp = currMB->cbp;
          }
        }

        // Go through transform modes. 
        // Note that if currMB->cbp is 0 one could choose to skip 8x8 mode
        // although this could be due to deadzoning decisions.
        //if (input->Transform8x8Mode==1 && currMB->cbp!=0) 
        if (input->Transform8x8Mode==1)
        {
          //=========== try mb_types 1,2,3 with 8x8 transform ===========
          if ((mode >= 1 && mode <= 3) && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try DIRECT-MODE with 8x8 transform ===========
          else if (mode == 0 && bslice && active_sps->direct_8x8_inference_flag && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try mb_type P8x8 for mode 4 with 4x4/8x8 transform ===========
          else if ((mode == P8x8) && (enc_mb.valid[4]) && (currMB->luma_transform_size_8x8_flag == 0))
          {
            currMB->luma_transform_size_8x8_flag = 1; //check 8x8 partition for transform size 8x8
            continue;
          }
          else
          {
            currMB->luma_transform_size_8x8_flag = 0;
            break;
          }
        }
        else
          break;
      }
    }

  }
}

/*!
*************************************************************************************
* \brief
*    Quantization RD decision process for the best mode
*************************************************************************************
*/
void ComputeCoeffRDCost(int mode, 
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short best_ctr16x16,
                          short best_cbp)
{
// perform Quantization rate - distortion optimization

  short   best_rdo=0, best_coeff=0;

  gpAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;
  gpAdaptiveQuantParam->dwCoeffIdx = 0;

  {
    if(img->AdaptiveQMOptFlag)
    {
      DecideBestRdoForMB(mode, currMB, enc_mb, min_rdcost, min_rate, i16mode, bslice, inter_skip, best_ctr16x16, &best_rdo);
      gpBestAdaptiveQuantParam->dwAdaptiveQPOIdx = gpAdaptiveQuantParam->dwAdaptiveQPOIdx = best_rdo;
    }

    if(img->AdaptiveCoeffFlag)
    {
      DecideBestCoeffForMB(mode, currMB, enc_mb, min_rdcost, min_rate, i16mode, bslice, inter_skip, best_ctr16x16, &best_coeff);
      gpBestAdaptiveQuantParam->dwCoeffIdx = gpAdaptiveQuantParam->dwCoeffIdx = best_coeff;
    }
  }

  if(gpBestAdaptiveQuantParam->dwAdaptiveCoeffFlag)
  {
    if(!gpBestAdaptiveQuantParam->dwCoeffIdx)
      input->disthres = 0;
    else
      input->disthres = 1;
  }

};

//------ for q_matrix.c
/*!
 ************************************************************************
 * \brief
 *    initialize modeling quantization parameters
 *
 ************************************************************************
 */
void InitScalingListParam (void)
{
  int i;
  for(i=0;i<5;i++)
  {
    SliceMatrixParam0[i] = 32;
    SliceMatrixParam1[i] = 16;
  }

  img->slice_scaling_model_param0[0]   = 32;
  img->slice_scaling_model_param1[0]   = 16;
  img->slice_scaling_model_param0[1]   = 32;
  img->slice_scaling_model_param1[1]   = 16;

  if(input->UseAdaptiveQuantMatrix)
  {
    gpModelingMatrixType->gdwMatrixGradient   = 32;
    gpModelingMatrixType->gdwMatrixDCgain     = 16;

    // create 1st quantization matrices 
    SetAutoScalingListData(0);

    if(input->UseAdaptiveQuantMatrix>0)
    {
      gpModelingMatrixType->gdwMatrixGradient= 32;
      gpModelingMatrixType->gdwMatrixDCgain  = 16;
      // create 2nd quantization matrices 
      SetAutoScalingListData(1);
    }
  }
}
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix
 *
 ************************************************************************
 */
void SetAutoScalingListData(int InternalFlag)
{

  int pA = 0, pB = 0,  pT = 0;

  pA = ((gpModelingMatrixType->gdwMatrixGradient)-32);
  pB = (gpModelingMatrixType->gdwMatrixDCgain);

  // 4x4 block
  CreateScalingList4x4(pT, pA, pB, InternalFlag);
  // 8x8 block
  CreateScalingList8x8(pT, pA, pB, InternalFlag);
}
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix at 4x4 block
 *
 ************************************************************************
 */
void CreateScalingList4x4(int pT, int pA, int pB, int InternalFlag)
{
  int i,j,r=0;
  unsigned int QM4[16];

  // 4x4 Matrix
  switch ( pT )
  {
    case 0:
      {
        for(j=0;j<4;j++){
          for(i=0;i<4;i++){
            r=i+j;
            QM4[i+(j<<2)]=Clip3(1,255,(((pA*r)+8)>>4) + pB);
          }
        }
      }
      break;
    default:
      {
        printf("ERROR\n");
        break;
      }
  }

  if(InternalFlag)
  {
    for(i=0;i<16;i++)
      for(j=0;j<6;j++)
        ScalingList4x4AQMS[j][i] = Clip3(1,255,QM4[i]);
  }
  else
  {
    for(i=0;i<16;i++)
      for(j=0;j<6;j++)
        ScalingList4x4[j][i] = Clip3(1,255,QM4[i]);
  }
}
/*!
 ************************************************************************
 * \brief
 *    For generating the modeling quantization matrix at 8x8 block
 *
 ************************************************************************
 */
void CreateScalingList8x8(int pT, int pA, int pB, int InternalFlag)
{
  int          i,j,r=0;
  unsigned int QM[64];

  // 8x8 Matrix
  switch ( pT )
  {
    case 0:
    {
      for(j=0;j<8;j++){
        for(i=0;i<8;i++){
          r=i+j;
          QM[i+j*8]=Clip3(1,255,( ((pA*r+16)>>5 ) + pB) );
        }
      }
      break;
    }
    default:
    {
      printf("ERROR\n");
      break;
    }
  }

  if(InternalFlag)
  {
    for(i=0;i<64;i++)
      for(j=0;j<2;j++)
      ScalingList8x8AQMS[j][i] = Clip3(1,255,QM[i]);
  }
  else
  {
    for(i=0;i<64;i++)
      for(j=0;j<2;j++)
        ScalingList8x8[j][i] = Clip3(1,255,QM[i]);
  }
}

#endif

#ifdef ADAPTIVE_QP
/*!
*************************************************************************************
* \brief
*    Quantization Parameter Cntrol for a macroblock
*************************************************************************************
*/
void DecideBestQPForMB(int mode, 
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short best_ctr16x16,
                          short *best_mode,
                          short *best_qp)
{
  const int cMaxQP = 2;
  int dwQPIdx, current_QP, checkQP;
  double cost_qp = *min_rdcost;

  gpAdaptiveQuantParam->dwCurrQP = current_QP = gdwBaseQp;

  for(dwQPIdx=-1;dwQPIdx<cMaxQP;dwQPIdx++)
  {
    // reset the optimal parameters
    gpAdaptiveQuantParam->dwAdaptiveQPOIdx = 0;
    gpAdaptiveQuantParam->dwQMatrixIdx = 0;
    gpAdaptiveQuantParam->dwCurrQP = checkQP = current_QP + dwQPIdx;
    currMB->qp = img->qp = gpAdaptiveQuantParam->dwCurrQP;
    set_chroma_qp(currMB);
    currMB->delta_qp = (currMB->qp - currMB->prev_qp);

    // High profile's limitation
    if( checkQP < -img->bitdepth_luma_qp_scale || checkQP > 51 ) continue;

//******************* re-set the P or B parameter ***********************************************************************************************************************
    PreSetBestCodingParameters(mode, currMB, enc_mb, bslice, inter_skip, best_ctr16x16);
//******************* decide best aqms_idx ***************************************************************************************************************************

    //--- transform size ---           
    currMB->luma_transform_size_8x8_flag = input->Transform8x8Mode==2
      ?  (mode >= 1 && mode <= 3)
      || (mode == 0 && bslice && active_sps->direct_8x8_inference_flag)
      || ((mode == P8x8) && (enc_mb.valid[4]))
      :  0;

    SetModesAndRefframeForBlocks (mode);
    // Encode with coefficients
    img->NoResidueDirect = 0;  

    {
      while(1)
      {
        if(RDCost_for_macroblocks (enc_mb.lambda_md, mode, &cost_qp, min_rate, i16mode))
        {
          if(cost_qp < *min_rdcost)
          {
            *best_qp = currMB->qp;
            *min_rdcost = cost_qp;
            store_macroblock_parameters (mode);
            gdwBestCbp = currMB->cbp;
          }
        }
        // Go through transform modes. 
        // Note that if currMB->cbp is 0 one could choose to skip 8x8 mode
        // although this could be due to deadzoning decisions.
        //if (input->Transform8x8Mode==1 && currMB->cbp!=0) 
        if (input->Transform8x8Mode==1)
        {
          //=========== try mb_types 1,2,3 with 8x8 transform ===========
          if ((mode >= 1 && mode <= 3) && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try DIRECT-MODE with 8x8 transform ===========
          else if (mode == 0 && bslice && active_sps->direct_8x8_inference_flag && currMB->luma_transform_size_8x8_flag == 0)
          {
            //try with 8x8 transform size
            currMB->luma_transform_size_8x8_flag = 1;
            continue;
          }
          //=========== try mb_type P8x8 for mode 4 with 4x4/8x8 transform ===========
          else if ((mode == P8x8) && (enc_mb.valid[4]) && (currMB->luma_transform_size_8x8_flag == 0))
          {
            currMB->luma_transform_size_8x8_flag = 1; //check 8x8 partition for transform size 8x8
            continue;
          }
          else
          {
            currMB->luma_transform_size_8x8_flag = 0;
            break;
          }
        }
        else
          break;
      }
    }

  }

}
#endif
